DOCKER_API = 1
FILE_NOT_FOUND = 2
KEYBOARD_INTERRUPT = 3
PARSER_ERROR = 4
PLUGIN_ERROR = 5
PLUGIN_EXCEPTION = 6
VARIABLE_ERROR = 7
TESTER_KEY_ERROR = 8
TEST_KEY_ERROR = 8
TRICOT_EXCEPTION = 9
TRICOT_RUNTIME_ERROR = 10
UNKNOWN = 11
VALIDATION_EXCEPTION = 12
VALIDATOR_ERROR = 13
VALUE_ERROR = 14
RUNTIME_ERROR = 15
CONDITION_FORMAT_ERROR = 16

LAST_ERROR = 0
